# bottalk.py -- "Bot talk"
#
# Copyright (C) 1999 Joel Rosdahl
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Joel Rosdahl <joel@rosdahl.net>
#
# $Id: bottalk.py,v 1.30 1999/09/29 21:31:55 joel Exp $

import locale
import re
import string
import time
import whrandom
import os

VERSION = 0, 0, 2

locale.setlocale(locale.LC_ALL, "en_GB.ISO8859-1")

nicknamecharsregexp = r"[]A-Za-z0-9[\`^{}-]"
nickidentifier = "%n"
substitutes = { "%words%" : "([\w ]*\w)",
                "%word%" : "(\w+)"
}

def findall(pattern, string):
    result = []
    r = re.compile(pattern)
    while string:
        m = r.search(string)
        if not m:
            return result
        if m.group():
            result.append(m.group())
        string = string[m.end():]
    return result

def klockan(s):
    return os.popen("date +%H.%M", "r").read()[:-1]

def matte(s):
    try:
        return str(eval(s))
    except ZeroDivisionError:
        return "odefinierat"
    except SyntaxError:
        return "inkorrekt"

class chatter:
    def __init__(self, nick,
                 file_base=".",
                 rules_file="talk.sentences",
                 rewrites_file="talk.rewrites",
                 synonyms_file="talk.synonyms",
                 sentences_base="sentences",
                 log_file="talk.log"):
        self.nick = string.lower(nick)
        self.rules = self.load_rules(file_base + "/" + rules_file)
        self.rewrites = self.load_rewrites(file_base + "/" + rewrites_file)
        self.synonyms = self.load_synonyms(file_base + "/" + synonyms_file)
        self.sentences_base = file_base + "/" + sentences_base
        self.log_file = file_base + "/" + log_file

    def answer(self, callernick, s):
        self.log('Answer "%s" from "%s" as "%s".' % (s, callernick, self.nick))
        if not self.should_answer(s):
            self.log("Shouldn't answer.")
            return []
        self.log("Should answer.")
        statements = self.fix_input_string(s)
        self.log("Statements: %s" % statements)
        answers = map(self.make_answer, len(statements)*[callernick], statements)
        # Remove empty answers.
        answers = filter(lambda x: x, answers)
        for i in range(len(answers)):
            self.log('Answered "%s"' % answers[i])
        return answers

    def log(self, str):
        l = open(self.log_file, "a")
        l.write(time.strftime("%c", time.localtime(time.time())) +
                " -- " + str + "\n")
        l.close()

    def load_rules(self, filename):
        f = open(filename, "r")
        lines = f.readlines()
        f.close()
        state = 0   # 0 = get priority, 1 = get match lines, 2 = get answers
        rules = []
        tmp_rule = []
        tmp_list = []
        for i in range(len(lines)):
            # Check for comment or blank lines
            if re.match(r"(#.*|\s+|$)", lines[i]):
                continue
            if lines[i] == "---\n":
                if state != 0:
                    tmp_rule.append(tmp_list)
                    tmp_list = []
                if state == 2:
                    rules.append(tmp_rule)
                    tmp_rule = []
                state = (state+1) % 3                
            else:
                lines[i] = lines[i][:-1] # remove trailing newline
                if state == 0:
                    tmp_rule = [int(lines[i])]
                else:
                    tmp_list.append(lines[i])
        return rules

    def load_rewrites(self, filename):
        f = open(filename)
        lines = f.readlines()
        f.close()
        rewrites = {}
        for i in range(len(lines)):
            # Check for comment or blank lines
            if re.match(r"(#.*|\s+|$)", lines[i]):
                continue
            tmp = string.split(lines[i][:-1], ":")
            rewrites[tmp[0]] = tmp[1]
        return rewrites
    
    def load_synonyms(self, filename):
        f = open(filename)
        lines = f.readlines()
        f.close()
        synonyms = {}
        for i in range(len(lines)):
            # Check for comment or blank lines
            if re.match(r"(#.*|\s+|$)", lines[i]):
                continue
            tmp = string.split(lines[i][:-1], ":")
            for j in range(1, len(tmp)):
                synonyms[tmp[j]] = tmp[0]
        return synonyms

    def should_answer(self, s):
##         # Always answer.  Boring, or...?
##         return 1
    
        s = strip_irc_codes(s)

        nick = re.escape(self.nick)

        r1 = re.search(re.escape(self.nick), s, re.IGNORECASE)
        if not r1:
            return 0
        r2 = re.search("\w", s)
        if r2 and (r1.start() == r2.start()):
            # Found nick at beginning of s.
            return 1

        r1 = re.match(".*(%s).*" % re.escape(self.nick), s, re.IGNORECASE)
        if not r1:
            return 0
        r2 = re.match(".*(\w).*", s)
        if r2 and (r1.end(1) == r2.end(1)):
            # Found nick at end of s.
            return 1
        return 0

    def fix_input_string(self, s):
        s = string.lower(s)
        s = string.replace(s, self.nick, "")
        s = re.match(r"[\W\s]*(.*)", s, re.LOCALE).group(1)
        list = findall(r"[^.!?]+\??", s)
        return map(string.strip, list)

    # If mapping == { "fie": "pie" } and string == "foo fie fum", then return
    # "foo pie fum"
    def substitute_string(self, string, mapping):
        if type(string) != type(""):
            return ""
        tokens = re.split(r"(\W)", string)
        result = ""
        for t in tokens:
            if mapping.has_key(t):
                result = result + mapping[t]
            else:
                result = result + t
        return result

    def substitute_synonyms(self, string):
        return self.substitute_string(string, self.synonyms)

    def substitute_rewrites(self, string):
        return self.substitute_string(string, self.rewrites)

    def make_answer(self, callernick, statement):
        self.log("make_answer(%s)" % statement)
        best_match = 0
        current_answer_list = []
        current_match_groups = ()
        
        statement = self.substitute_synonyms(statement)

        # Loop over all rules.
        for i in range(len(self.rules)):
            current_rule = self.rules[i]
            if current_rule[0] <= best_match:
                continue

            # Try to find a match for the current rule.
            for j in range(len(current_rule[1])):
                rule = current_rule[1][j]
                the_rules = re.split(":", rule)
                for k in substitutes.keys():
                    for l in range(len(the_rules)):
                        the_rules[l] = string.replace(the_rules[l], k, substitutes[k])

                # Create a match object for the first rule.  I.e., the
                # group information is taken from the first rule only.
                m = re.search(the_rules[0], statement, re.LOCALE|re.IGNORECASE) 
                if m:
                    # It matched! Now check that the other rules match.
                    rules_match = 1
                    for k in range(1, len(the_rules)):
                        if the_rules[k][0] == '!':
                            # inverted rule
                            if re.search(the_rules[k][1:], statement, re.LOCALE|re.IGNORECASE):
                                rules_match = 0
                                break
                        else:
                            # normal rule
                            if not re.search(the_rules[k], statement, re.LOCALE|re.IGNORECASE):
                                rules_match = 0
                                break

                    if rules_match:
                        # Yeah!  Got a match for all rules!
                        self.log('Found match: "%s" (score: %d, best match: %d)' % (the_rules,
                                                                                    current_rule[0],
                                                                                    best_match))
                        best_match = current_rule[0]
                        current_answer_list = current_rule[2]
                        current_match_groups = []
                        # Fix "mina" --> "dina" (et c) in match groups.
                        for g in m.groups():
                            current_match_groups.append(self.substitute_rewrites(g))
                        break

        if current_answer_list:
            answer = whrandom.choice(current_answer_list)
            if answer == "svara inte":
                self.log("svarade inte...")
                return ""

            answer = self.expand_answer(answer)

            # Substitute %1, %2, ...
            for i in range(len(current_match_groups)):
                answer = string.replace(answer, "%" + str(i+1), current_match_groups[i])

            # Evaluate something?
            m = re.match("(.*)%eval ([^(]*)\((.*)\)%(.*)", answer)
            if m:
                answer = (m.group(1) +
                          eval("%s('%s')" % (m.group(2), m.group(3))) +
                          m.group(4))

            # Maybe capitalize first character...
            if whrandom.randint(0, 1) == 0:
                answer = string.upper(answer[0]) + answer[1:]

            # Add attribution for the caller.
            if string.find(answer, nickidentifier) == -1:
                answer = "%s: %s" % (callernick, answer)
            else:
                answer = string.replace(answer, nickidentifier, callernick)

            # Expand character at end...
            answer = self.expand_character_at_end(".", answer)
            answer = self.expand_character_at_end("!", answer)
            return answer
        else:
            return ""

    # Substitute character "c" in the end to "", "c", "cc", or "ccc".
    def expand_character_at_end(self, char, string):
        if string[-1] == char:
            return string[:-1] + char*whrandom.randint(0, 3)
        else:
            return string
        

    def expand_answer(self, string):
        while 1:
            match = re.search("%\w+%", string, re.LOCALE)
            if not match:
                return string
            string = (string[:match.start()] +
                      self.lookup_expansion(match.group()) +
                      string[match.end():])

    def lookup_expansion(self, string):
        self.log("lookup_expansion(%s)" % string)
        string = string[1:-1]   # %word% --> word
        try:
            f = open(self.sentences_base + "/" + string)
        except IOError:
            self.log("Couldn't find sentences/%s." % string)
            return "fel, ju"
            
        answers = f.readlines()
        f.close()
        return self.expand_answer(whrandom.choice(answers)[:-1])

def strip_irc_codes(s):
    return re.sub("(|||\d{1,2}(,\d{1,2})?)", "", s)
